'************************************************************
' Quad NIMH Discharger V0.90 (beta)
' Capacity limits ~1800-3000mAh / 1.8R
' MMBasic 4.5 / TFT-Maximite
' by twofingers 09-2014 at TBS
'-------------------------------------------------
' Purpose:
' This program discharges up to 4 rechargeable
' NiMH batteries to measure the capacity.
' Results (mAh) and the discharge curve are shown on
' the VGA and/or mAh only on a 16x2 I2C LCD (if
' there is any). LCD support is still "basic".
' You can store a screenshot and a CSV file.
'
' Ideas for possible improvements/changes:
' 1. PWM control of discharge current
' 2. storing log buffer to eeprom
' 3. a charger
' 4. a uMite version (a conversion sould be easy)
' 5. Autostart w. count down?
' 6. ...
'    ... or make a data logger for temperatur or similar
'

' Note/Warning:
' Frequent discharge can/will
' shorten rechargeable batteries life!
' Any cell has a limited lifecycle.
'
'-------------------------------------------------
'
' Used GPIO PINs:
' 1-4   as analog input
' 14-17 as digital output
'
'-------------------------------------------------
' Used special keys:
' ENTER  starts the discharge (if there any batteries)
' SPACE  aborts the discharge
' [F1]   changes the filename for BMP and CSV-file. Must
'        be used before(!) entering the discharge mode.
'        If filename = "________" then nothing will be saved
' [F2]   toggles the store mode ON/OFF
'        can be used in discharge mode
' Ctrl-C to quit the program
'-------------------------------------------------
' After discharging two files will be generated:
' 1. a vga screenshot as ".BMP" file
' 2. a ".CSV" file w. Time, Date, mAh
'    and a serie of voltage values,
'    each for a 15 sec period.
'-------------------------------------------------
' Hardware (almost nothing):
' a (simple, used?) battery charger case w. 4 slots
' for each slot:
' 2x 100k, 1x 1.8R/2W, 1x IRL540N (logic level N-MOSFET)
' 1x perfboard, some small cables,
' 1x flat ribbon cable w. 26p GPIO connector
'    (ie old PC printer MB adaptor)
'-------------------------------------------------
'
'    A IN (Pin 1-4)
'   o---------------------o---------------o----------+
'                         |               |          |
'                        .-.              |          |
'                        | |              |          |
' << to Maximite         | | 1.8R         |          |
'                        '-'              |          ^ Test
'                         |               |          | NiMH
'                         o              .-.         | Battery
'   D OUT (Pin 14-17)     o              | | 100k    |
'   o--------o------+     |              | |        ---
'            |      |  ||-|              '-'         -
'            |      |  ||<) IRL540        |          |
'           .-.     +--||-|               |          |
'           | |           |               |          v
'      100k | |           |               |          |
'           '-'           |               |          |
'            |            |               |          |
'   o--------o------------o---------------o----------+
'    Gnd
'************************************************************
' This code may be freely distributed and
' changed. Provided AS IS without any warranty.
' Use it at your own risk.
' ---------------------------------------------
' I have to give credit to Peter (G8JCF)
' for his very useful LCD I2C library
' and of course to Geoff for his great MMBasic
'************************************************************
'************************************************************


Option error continue       ' continue after a I/O error

' for TFT-Maximite the 1602 LCD is disabled
'Library Load "lcdi2cli.bas" 'G8JCF's LCD library

'Lcdi2c_init &H27,1,0,2,3
'Lcdi2c_clear
'Lcdi2c_backlight 1

Dim v(4) 'for voltage
Dim  mAh(4),slot_used(4),accu_empty(4),t(4)

Dim logbuf$(4,1200) length 1   ' log buffer for 5h x 240samples (4s/min)
Dim logbuf_ptr(4),plot_pos(4)
'Dim p(4) 'Lcd_alive
Dim HIGH,LOW,TRUE,FALSE
Dim fname$, bmpfname$, busy_Slots, userStop
Dim bPlural$(1),fill$

' Const
HIGH=1
LOW=0
TRUE=1
FALSE=0
progVer$="0.90"

'OFF=0   'predefined
'AIN=1   'predefined
'DOUT=8  'predefined

                ' please read the batterie manual for specific values
threshold=1.00  ' discharge limit, final voltage
                ' 1.00 Volt is default value for most NiMH Batteries
                ' should never be below 0.90 V
maxDischargeTime=5*3600*1000 ' 5h, about >3000mAh w. 1.8R
' Note: V_offset=0.03 'Spannungsverlust ueber MOSFET

' Vars
bmpFNAME$="BATFILE3" ' default file name for BMP, "________" = will not save
save_on=TRUE         ' toggle with [F2]

light=TRUE           ' for LCD and flashing LED (PIN(0))
Dim reverse          ' print reverse, b/w

bPlural$(0)="y"
bPlural$(1)="ies"
fill$      =""

Touch Size 76,84
hOff=36          'touch button horizontal offset

Pause 300  ' important for keydown!!

'*********************************
Do '  The endless Mainloop
'*********************************
  userStop =FALSE
  readyFlag=FALSE
  notouch =TRUE   ' Touch button not visible

  PrepScreen
  'LCDI2C_Clear

  ' Initialisation - look out for batteries
  For i = 1 To 4
    t(i)=0          ' time
    v(i)=0          ' voltage
    mAh(i)=0        ' capacity
    plot_Pos(i)=0   ' graphic pixel output
    logbuf_ptr(i)=0 ' points @ current log buffer pos
    SetPin i,AIN    ' analog input for battery voltage
    accu_Empty(i)=FALSE
  Next i

  Do ' is there something to do?
     ' (Note: If you want to test this program without the
     ' required hardware you can set busy_Slots=1 here)
    busy_Slots=0
    Print @(6*68,12*20) Time$
    Print @(6*68,12*21) Date$

    For i = 1 To 4
      slot_Used(i)=FALSE
      v(i)=Pin(i)
      CSet 1+33*Not (i And 1),2+11*(i>2)
      volt$=Format$(v(i),"%2.2f")
      Print "Slot"+Str$(i)+": ";
      Color Green
      Font#1,,v(i)>0: Print volt$+"V ":      Font#1,,0
      Color White
      'LCDI2C 1+(i>2),1+Not (i And 1)*8, Str$(i)+":"+volt$
      If v(i)>threshold Then
        slot_Used(i)=TRUE
        busy_Slots=busy_slots+1
      EndIf
    Next i

    If busy_Slots>0 Then
      If notouch Then
         notouch =FALSE
         Touch Create 1, 403,hOff, "Start",Green,b
         Clr 12
      EndIf
      Font#1,,1
      Center 12, "   Press ENTER to discharge    "
      Font#1
      Cset(67,12):Print "[F1] "bmpFNAME$;
      keyVal=KeyDown
      If keyVal = 145 Then 'F1 pressed? edit filename
        Clr 12
        Cset(72,12)
        bmpFNAME$= LineEdit$(bmpFNAME$)
        Cset(67,12):Print "     "+bmpFNAME$;
            Print @(0,0) keyval
      EndIf
      If keyVal = 13 Or TouchVal(1) Then
         Cset(67,12):Print "[F2] ";
         Touch remove 1
         readyFlag=TRUE
      EndIf
    Else
      notouch =TRUE
      Touch remove 1
      Color Green
      arrow_p=(arrow_p+1) And &B111
      Center 12, " >>"+Space$(8-arrow_p)+"Please insert your battteries"+Space$(8-arrow_p)+"<< "
      Cset(67,12):Print "    ";        ' delete "[F1]"
      Pause 40
      Color White
    EndIf
  Loop While Not readyFlag

  '**  there is somthing to do!

  'new ctrl-C handling, important to close the gates (MOSFET)
  On KEY New_CtrlC
  Option BREAK 255

  For i = 1 To 4     ' open the gates, lets the force flow through!
    If v(i)>threshold Then
      SetPin i+13,DOUT
      Pin(i+13)=HIGH
    EndIf
  Next i

  Timer=0
  ticker=0:ex_ticker=0
  SetTick 1000,Tick_handler

  '*************************
  '     Discharge Loop
  '*************************
  Do
   Touch Create 1, 403,hOff, "ABORT",Red,b
   Font#1,,reverse

   bP$=Str$(busy_Slots)+" batter"+bPlural$(busy_Slots>1)
   fill$=Space$(busy_Slots=1) ' one chr$(32) if there only 1 battery leftCenter 12," Discharging "+battno$+" - press <space> to abort "
   Center 12,fill$+" Discharging "+bP$+" - press <space> to abort "+fill$
   Font#1
   reverse=Not reverse  ' inverse letters, blinking msg

   For i = 1 To 4
    If slot_Used(i) Then
      If accu_Empty(i)=FALSE Then
        v(i)=Pin(i)    ' get the voltage of slot i
'       'p(i)=Not p(i) ' still alive marker for LCD
        If (Timer/1000 Mod 15) = 0 Then '--- a yet unused program part --
          logbuf_ptr(i)=logbuf_ptr(i)+1        ' here we store
          logbuf$(i,logbuf_ptr)=Chr$(v(i)*100) ' the voltage in an array
        EndIf                           '---------------------------------

        If v(i)=<threshold Then ' close this gate
          Pin(i+13)=LOW
          SetPin i+13,OFF
          SetPin i,OFF
          accu_Empty(i)=TRUE
          busy_Slots=busy_Slots-1
          'Lcd_result i
          Result i: Print" OK"
        Else
          Plot plot_pos(i), i
          If (Timer/1000 Mod 90) = 0 Then ' 90= 60*60/40= 40 dots/h
             plot_pos(i)=plot_pos(i)+1
          EndIf
          t(i)= Timer
          mah(i)=mah(i)+v(i)/1.8/3.600 'mAh = V/R/3600sec*1000
          'Lcd_result i
          Result i
        EndIf
      Else
        'Lcd_result i
        Result i:Print " OK"
      EndIf
    Else
      'Lcd_result i
      CSet 1+33*Not (i And 1),2+11*(i>2)
      Print "Slot"Str$(i)": ";"- Empty or no contact -"
    EndIf
   Next i

   Do:Loop While ticker = ex_ticker ' wait until 1 sec has passed
   ex_ticker=ex_ticker+1
   If TouchVal(1) Then userStop=TRUE
  Loop While busy_slots>0 And Timer<maxDischargeTime And (Not userStop)
  '********************
  ' End Discharge Loop
  '********************
  Touch remove 1

  ' Housekeeping
  If busy_slots Then ' user stop
    Clr 12
    Center 12, "Resetting PINs ..."
    For i = 1 To 4
      SetPin i,OFF
      SetPin i+13,OFF
      Pin(i+13)=LOW
    Next i
    Pause 1000
  EndIf
                  ' now we can
  Option BREAK 3  ' enable regular CTRL-C function
  On KEY 0      ' and key handler

  If bmpFNAME$<>"________" And save_on Then
    Clr 12
    Cset(25,12): Print "Writing screenshot to SD ...         Savefile: "+bmpFNAME$;
    InvertScreen      ' makes printing easier
    SaveBMP bmpFNAME$
    InvertScreen
    If mm.error>0 Then 'File number ist not open
      Clr 12
      Center 12, "File Error (BMP):"+mm.error+"! Press any key to continue!"
      Waitkey
    EndIf

    Open bmpFNAME$+".CSV" For output As 1
    If mm.error=0 Then
      Print #1,"NiMH Battery Discharger File "+progVer$+" "+Date$+" "+Time$
      For i = 1 To 4
        Print #1, i","logbuf_ptr(i)", "+Format$(mAh(i),"%4.0f")+",mAh,";
        For j = 1 To logbuf_ptr(i)
           Print #1,Format$(Asc(logbuf$(i,j))/100,"%1.2f")+",";
        Next j
        Print #1
        If mm.error>0 Then 'File number ist not open
          Clr 12
          Center 12, "File Error (CSV1):"+mm.error+"! Press any key to continue!"
          Waitkey
        EndIf
      Next i
    Else
      Clr 12
      Center 35, "File Error (CSV2):"+mm.error+"! Press any key to continue!"
      Waitkey
    EndIf
    Close #1
    Pause 200 ' let's see the messages

  EndIf
  Clr 12
  Center 12, "DONE! Press any key to restart!"
  If save_on Then Cset(62,12):Print "Savefile: "+bmpFNAME$;

  Touch Create 1, 403,hOff, "DONE",Blue,b

  Light=1
  Do
    'lcdi2c_clear
    'lcdi2c 1,6,"Ready!"
    'lcdi2c 2,2,"Press any Key!"
    Pause 1000
    Light=Not Light
    'LCDI2C_Backlight Light

    Pause 300
    Light=Not Light
    'lcdi2c_clear
    'LCDI2C_Backlight Light

    For i = 1 To 4
     ' lcd_result i
    Next i

    Pause 1000
  Loop While Inkey$="" And Not TouchVal(1)
Loop '***** END of Mainloop *********
End  ' will never be reached
'************************************


Sub Tick_handler ' inc ticker once a second
  ticker=ticker+1
End Sub


Sub Lcd_Result i ' Print the discharge result on LCD screen
   Local m$(2)
   m$(0)="^":m$(1)=" " 'still alive marker char

   If slot_used(i) Then
     msg$=Format$(i,"%g")+":"+Format$(mah(i),"%4.0f")
     If v(i)=<threshold Then
        MSG$=MSG$+"*"
     Else
        MSG$=MSG$+m$(p(i))
     EndIf
   Else
     msg$=Format$(i,"%g")+":Empty"
   EndIf

   Lcdi2c 1+(i>2),1+Not(i And 1)*8,msg$ '1=1,1: 2=1,9: 3=2,1: 4=2,9
End Sub


Sub Result i ' Print the discharge result on main screen
   Cset(67,12)
   If save_on Then
      Print "[F2] "bmpFName$;
   Else
      Print "SAVE FILE=";
      Font#1,,1
      Print "OFF";
      Font#1
   EndIf
   Print @(6*68,12*20) Time$;
   Print @(6*68,12*21) Date$;
   CSet 1+33*Not (i And 1),2+11*(i>2)
   Print "Slot"+Str$(i)+": ";
   Print Format$(v(i),"%2.2f")"V ";
   Print Format$(mah(i),"%4.0f")"mAh ";
   Print Sec2Hms$(Int(t(i)/1000));
End Sub


Function Sec2Hms$ sec ' convert sec to "00:00:00" string
                      ' limit = 4,17 days (359999 sec)
  Local Hour,Mins,Secs

  hms.days= Int(sec/86400)  ' just if you need to know
  Hour= Int(sec/3600)
  Mins= Int((sec Mod 3600)/60)
  Secs= Int((sec Mod 3600) Mod 60)
  Sec2Hms$=Format$(Hour,"%02g")+":"+Format$(Mins,"%02g")+":"+Format$(Secs,"%02g")
End Function


Sub New_CtrlC ' my CTRL-C handler
Local key_value, i
  key_value=KeyDown

  If key_value=3 Then 'CTRL-C
    Clr 12
    Center 12,"Resetting PINs ..."
    For i = 1 To 4
      SetPin i,OFF
      SetPin i+10,OFF
      Pin(i+10)=LOW
    Next i
    Option BREAK 3  ' re-enable regular CTRL-C function
    End
  EndIf

  If key_value=32 Then ' space bar means stop
    Clr 12
    Center 12,"USER STOP"
    UserStop= TRUE
    Pause 500
  EndIf

  If key_value=146 Then  ' F2
    save_on= Not save_on ' toggle save bmp
  EndIf
End Sub


Sub Waitkey ' just wait for (almost) any key
  Do: Loop While Inkey$=""
End Sub


Sub CSet col, row ' set print cursor pos, col=0-79,row=0-35
  Print @(col*6,row*12);
End Sub


Sub Center row, msg$,offset ' print a message at row
  Print @(((MM.HRes/6-Len(msg$))/2)*6,row*12+offset)msg$;
End Sub


Sub Plot a,i ' a=plot position, i=slot
  Pixel(a+(MM.HRes-80)/2*(Not(i And 1)),132*(i>2)+200-v(i)*100)=Yellow
End Sub


Sub Clr row,color ' clear row
  Line (0,row*12) - (MM.HRes,row*12+12),color,BF
End Sub


Sub InvertScreen '
  Line (0,0)-(MM.HRes-1,MM.VRes-1),-1,bf
End Sub


Sub PrepScreen ' prepare screen main screen
Local vt, vb
  Cls
  Color White
  Line(0,4)-(478,4)
  Center 0,"<< TFT-MAXIMITE Quad NiMH Discharger V0.90 >>"

  cset(72,12):Print bmpFNAME$;

  vt=120    ' baseline top, slot 1+2
  vb=240+12 ' baseline bottom, slot 3+4

  ' Axes
  Line (0,vt)-(400,vt)    'horizontal - top
  Line (0,vb)-(400,vb)    'horizontal - bottom
  Line (200,36) -(200,vt) 'vertikal - top

  ' H-Scale bottom -----------------------
  For i = 0 To 400 Step 40
    Line (i, 36)-(i,vt) 'vertikal - top
    Line (i,168)-(i,vb) 'vertikal - bottom
  Next i

  ' H-Scale top --------------------------
  For i = 0 To 400 Step 40
    Line (i,vt-5)-(i,vt+5) 'vertikal - top
    Line (i,vb-5)-(i,vb+5) 'vertikal - bottom
  Next i

  ' V-Scale bottom -----------------------
  For i = vb To 168 Step -10
    Line (0,i)-(2,i) 'horizontal - bottom
  Next i
  Line (0,vb-100+80)-(5,vb-100+80) ' 1V

  For i = vb To 168 Step -10
    Line (198,i)-(202,i) 'horizontal - bottom
  Next i
  Line (195,vb-100+80)-(205,vb-100+80) ' 1V
  Circle(0,vb-threshold*100+80),2,Red,f
  Circle(200,vb-threshold*100+80),2,Red,f

  ' V-Scale top --------------------------
  For i = vt To 36 Step -10
    Line (0,i)-(2,i) 'horizontal - top
  Next i
  Line (0,vt-100+80)-(5,vt-100+80) ' 1V

  For i = vt To 36 Step -10
    Line (198,i)-(202,i) 'horizontal - top
  Next i
  Line (195,vt-100+80)-(205,vt-100+80) ' 1V
  Circle(0,vt-threshold*100+80),2,Red,F
  Circle(200,vt-threshold*100+80),2,Red,F

  ' Labels
  For i = 1To 4
    p= (i*40)-12

    Print @(p    ,22*12-6)i"h";
    Print @(p+200,22*12-6)i"h";
    Print @(p    ,11*12-6)i"h";
    Print @(p+200,11*12-6)i"h";
  Next i
  Print @(p+40-8,22*12-6)i"h|0h";
  Print @(p+40-8,11*12-6)i"h|0h";
End Sub '******* PrepScreen


'***********************************************
'*
'*                LineEdit$
'*
'*          MMBasic 4.5/Maximite
'*   w. modifications for legal filenames
'*  (you can find a regular version on TBS)
'***********************************************
Function LineEdit$ edText$
Local hpos, vpos, p, ep, keyvalue, enterPressed
Local text$,ltext$,mtext$,rtext$     ' mtext$=cursor
Local TRUE, FALSE
'Local insertMode

  TRUE=1
  FALSE=0
  hpos=MM.HPos:vpos=MM.VPos ' save print position
  enterPressed=FALSE        ' Flag to EXIT the loop
  text$=edText$             ' lets use a local copy of edText$
  LineEdit$=edtext$         ' backup text$ for Esc
  p=Len(text$)              ' cursor pointer
  ep=p                      ' end pointer, len of edText$

  Pause 200 ' es gibt einen Zeitfaktor fuer keyboard input buffer ~140ms

  Print @(hpos,vpos)Left$(text$,p-1);
  Font#1,,1:Print Right$(text$,1);:Font#1

  Do
    Do:keyvalue=KeyDown:Pause 20:Loop While keyvalue=0

    If keyvalue=130 And p>1 Then p=p-1    'left arrow
    If keyvalue=131 And p<ep Then p=p+1   'right arrow
    If keyvalue=13 Then
       Print @(hpos,vpos)LineEdit$;
       enterPressed=TRUE                  'Enter
    EndIf
    If keyvalue=27 Then
       Print @(hpos,vpos)LineEdit$;
       Exit Function                      'ESC
    EndIf
    If keyvalue=134 Then p=1              'Home
    If keyvalue=135 Then p=ep             'End
'>>>If keyvalue=132 Then insertMode=Not insertMode
                                          'Insert toggle (still dummy)
    If keyvalue>32 And keyvalue<127 Then 'all ASCII chars, no blanks please
      If p<ep Then
        p=p+1
        ltext$=ltext$+Chr$(keyvalue)
        mtext$=Mid$(text$,p,1)
        rtext$=Mid$(text$,p+1)
      Else
        ltext$=Left$(text$,p-1)
        mtext$=Chr$(keyvalue)
        rtext$=""
      EndIf
    ElseIf keyvalue=8 And p>1 Then        'Backspace, no blanks please
      p=p-1
      ltext$=Mid$(text$,1,p-1)
      mtext$="_" 'blank replacement
      rtext$=Mid$(text$,p+1)
    ElseIf keyvalue=127 Then              'Del, no blanks please
      ltext$=Mid$(text$,1,p-1)
      mtext$="_" 'blank replacement
      rtext$=Mid$(text$,p+1)
    Else
      ltext$=Mid$(text$,1,p-1)            'move cursor
      mtext$=Mid$(text$,p,1)
      rtext$=Mid$(text$,p+1,ep-p)
    EndIf

    Print @(hpos,vpos)ltext$;
    Font#1,,1
    Print mtext$;
    Font#1
    Print rtext$;

    text$=ltext$+mtext$+rtext$

    Pause 140 ' some wait states to slow down key input
  Loop While Not enterPressed
  LineEdit$=ltext$+mtext$+rtext$
End Function
'*****************************************************
